use std::time::Duration;

use anyhow::{anyhow, Context};
use config::Config;
use lazy_static::lazy_static;
use log::{debug, error, info};
use rand::{distributions::Standard, seq::IteratorRandom, Rng};
use shows::Shows;

mod config;
mod shows;
mod video;

use crate::shows::EpisodeNumber;

lazy_static! {
    static ref RETRY_INTERVAL: Duration = Duration::from_secs(30);
}

#[tokio::main]
async fn main() -> anyhow::Result<()> {
    dotenvy::dotenv().ok();
    env_logger::init();
    ffmpeg_next::init().unwrap();

    let mut rng = rand::thread_rng();

    let conf = config::load();

    info!("Loading shows from {}", conf.shows_file.display());
    let shows = shows::load(&conf.shows_file).with_context(|| {
        format!(
            "Failed to load shows from file {}",
            conf.shows_file.display()
        )
    })?;
    if shows.is_empty() {
        return Err(anyhow!("Shows file is empty!"));
    }

    info!("Logging into cohost as {}", conf.cohost_email);
    let session = eggbug::Session::login(&conf.cohost_email, &conf.cohost_password)
        .await
        .context("Failed to login to cohost")?;

    loop {
        let result = post_random_screencap(&conf, &shows, &session, &mut rng)
            .await
            .context("Failed to post a random screencap");

        if conf.post_interval == Duration::ZERO {
            return result;
        }

        let delay = match result {
            Ok(()) => conf.post_interval,
            Err(e) => {
                error!("{}", e);
                *RETRY_INTERVAL
            }
        };
        tokio::time::sleep(delay).await;
    }
}

async fn post_random_screencap<R: Rng>(
    conf: &Config,
    shows: &Shows,
    session: &eggbug::Session,
    rng: &mut R,
) -> anyhow::Result<()> {
    let (title, show) = shows.iter().choose(rng).unwrap();
    let episodes = show.episodes().with_context(|| {
        format!(
            "Failed to get episode list for show {} with path {}",
            title,
            show.path.display()
        )
    })?;
    let (num, file) = episodes.iter().choose(rng).unwrap();

    let descriptor = format!(
        "{}{}",
        title,
        match num {
            EpisodeNumber::Standalone => String::new(),
            EpisodeNumber::SingleSeason(n) => format!(" episode {}", n),
            EpisodeNumber::MultiSeason(season, ep) => format!(" season {} episode {}", season, ep),
        }
    );

    info!("Selected: {} - {}", descriptor, file.display());

    let video_info = video::get_video_info(file, Some("eng")).with_context(|| {
        format!(
            "Failed to get duration and subtitle stream index for video file {}",
            file.display()
        )
    })?;
    debug!(
        "Video duration: {}",
        format_timestamp(video_info.duration_secs, None)
    );
    debug!(
        "Subtitle stream index: {:?}",
        video_info.subtitle_stream_index
    );

    let timestamp = video_info.duration_secs * rng.sample::<f64, _>(Standard);
    let formatted_timestamp = format_timestamp(timestamp, Some(video_info.duration_secs));
    info!("Taking screencap at {}", formatted_timestamp);

    let img_data = video::take_screencap(file, timestamp, video_info.subtitle_stream_index)
        .await
        .context("Failed to take screencap")?;

    let img_size = imagesize::blob_size(&img_data)
        .context("Failed to get image size for screencap image data")?;

    let attachment = eggbug::Attachment::new(
        img_data,
        format!("{} @{}.png", descriptor, formatted_timestamp),
        String::from("image/png"),
        Some(img_size.width as u32),
        Some(img_size.height as u32),
    )
    .with_alt_text(format!(
        "Screencap of {} at {}",
        descriptor, formatted_timestamp
    ));

    let mut tags = show.tags.clone();
    tags.extend_from_slice(&conf.global_tags);

    let chost_id = session
        .create_post(
            &conf.cohost_page,
            &mut eggbug::Post {
                content_warnings: vec![descriptor],
                attachments: vec![attachment],
                tags,
                draft: false,
                adult_content: false,
                headline: String::new(),
                markdown: String::new(),
                metadata: None,
            },
        )
        .await
        .context("Failed to create chost")?;
    info!("Created post {}", chost_id);

    Ok(())
}

fn format_timestamp(timestamp: f64, total_duration: Option<f64>) -> String {
    let total_duration = total_duration.unwrap_or(timestamp);
    format!(
        "{}{:02}:{:05.2}",
        if total_duration >= 3600.0 {
            format!("{}:", (timestamp / 3600.0) as u32)
        } else {
            String::new()
        },
        ((timestamp % 3600.0) / 60.0) as u32,
        timestamp % 60.0
    )
}
