use std::{
    collections::HashMap,
    fs,
    path::{Path, PathBuf},
};

use anyhow::{anyhow, Context};
use log::{debug, error};
use serde::Deserialize;

mod enumeration;

#[derive(Deserialize)]
pub struct Show {
    pub path: PathBuf,
    #[serde(default)]
    pub tags: Vec<String>,
}

pub type Shows = HashMap<String, Show>;

#[derive(Debug, Eq, Hash, PartialEq)]
pub enum EpisodeNumber {
    Standalone,
    SingleSeason(u32),
    MultiSeason(u32, u32),
}

type Episodes = HashMap<EpisodeNumber, PathBuf>;

pub fn load<P: AsRef<Path>>(shows_file: P) -> anyhow::Result<Shows> {
    serde_yaml::from_reader(fs::File::open(shows_file).context("Failed to open shows file")?)
        .context("Failed to parse YAML from shows file")
}

impl Show {
    pub fn episodes(&self) -> anyhow::Result<Episodes> {
        let path = &self.path;
        let metadata =
            fs::metadata(path).context("Failed to stat the show's path to determine file type")?;
        if metadata.is_file() {
            debug!("{} is a file, standalone", path.display());
            Ok(HashMap::from([(
                EpisodeNumber::Standalone,
                path.to_path_buf(),
            )]))
        } else if metadata.is_dir() {
            debug!("{} is a directory, enumerating episodes", path.display());
            let files: Vec<PathBuf> = fs::read_dir(path)
                .context("Failed to get a directory listing for the show's path")?
                .map(|entry| {
                    let entry =
                        entry.context("Failed to read a directory entry under the show's path")?;
                    if !entry
                        .file_type()
                        .with_context(|| {
                            format!(
                                "Failed to get file type for directory entry {}",
                                entry.path().display()
                            )
                        })?
                        .is_file()
                    {
                        debug!("Skipping {}, not a file", entry.path().display());
                        return Ok(None);
                    }
                    if entry.file_name().into_string().is_err() {
                        error!(
                            "Path {} contains invalid unicode, skipping",
                            entry.path().display()
                        );
                        return Ok(None);
                    }
                    Ok(Some(entry.path()))
                })
                .filter_map(|r| r.transpose())
                .collect::<anyhow::Result<Vec<PathBuf>>>()?;
            enumeration::enumerate_episodes(files)
                .ok_or(anyhow!("Could not detect any episode numbering scheme"))
        } else {
            Err(anyhow!("The show's path is not a file or a directory"))
        }
    }
}
